#include <sys/poll.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <limits.h>
#include <fcntl.h>
#include <sched.h>
#include <assert.h>
#include <stdint.h>

#include "common.h"
#include "macros.h"
#include "xmalloc.h"
#include "signal-event.h"
#include "io.h"
#include "misc.h"
#include "timeout.h"
#include "failures.h"
#include "vm.h"
#include "sock-unix.h"
#include "sock-inet.h"
#include "vmconf.h"
#include "task.h"
#include "sched.h"
#include "hal-proto.h"
#include "hal-runtime.h"
#include "bo.h"

//#define DEBUG_VMACHINED

#define warn(fmt, arg...) fprintf(stderr, "warn: %s: "fmt CRLF, \
					__FUNCTION__, ##arg);
#define info(fmt, arg...) fprintf(stderr, "info: %s: "fmt CRLF, \
					__FUNCTION__, ##arg);

#define error(eval, fmt, arg...)			\
do {							\
	fprintf(stderr, "error: %s: "fmt CRLF,		\
				__FUNCTION__, ##arg);	\
	exit(eval);					\
} while (0)


#ifdef DEBUG_VMACHINED
#define debug(fmt, arg...) fprintf(stderr, "debug: %s: "fmt CRLF, \
					__FUNCTION__, ##arg);
#else                                                                                                                       
#define debug(fmt, arg...) do {} while(0) 
#endif

#define	VMACHINED_RT_PRIORITY	50

/*
 * For interrupts we use 32 bits variable.
 * 0 is the highest interrupt priority.
 * 31 is the lowest interrupt priority.
 */
#define INTR_NUM_MIN	0
#define INTR_NUM_MAX	31

/*
 * Interrupts are generated from outside sources via a realtime signal.
 * Cycle task is started by a realtime timer.
 */
#define SIGINTERRUPT	SIGRTMIN
#define SIGCYCLE	(SIGRTMIN + 1)

#define CYCLE_TIMEOUT_SEC_DEFAULT	1

static struct {
	uint32_t	iflag;		/* raise bit when interrupt comes */
	uint32_t	ilevel;	/* mask for the current interrupt task */
} interrupt_state;

static struct {
	unsigned long	cycle_usec_user;	/* user defined cycle limit */
	unsigned long	cycle_usec_last;	/* last cycle run time */
	unsigned long	cycle_usec_min;		/* min cycle run time */
	unsigned long	cycle_usec_max;		/* max cycle run time */
	unsigned long	cycle_run;		/* count cycle's number */
} vmachined_stats;

extern char		*__progname;
static unsigned int	run_cycle_task;

/* All interrupt tasks are stored in the list. */
static struct task	*inter_list;
static struct task	*cycle_task, *bg_task;

static struct io_context *io;
static struct vm	*machine;

/* Pointers to machine images and their sizes. */
static unsigned char	*inputs_image;
static unsigned int	inputs_image_sz;
static unsigned char	*outputs_image;
static unsigned int	outputs_image_sz;
static int		byte_order;

/* Cycle realtime timer. */
static timer_t		cycle_timerid;

/* Cycle timeout in seconds and nanoseconds. */
static long		cycle_timeout_sec;
static long		cycle_timeout_nsec;

/* Cycle statistic timers. */
static struct timeval	tv_begin, tv_end;

static void cycle_timer_arm();
static void cycle_timer_disarm();
static void sig_block(int signo);
static void sig_unblock(int signo);

/* Handlers for interrupt and cycle signals. */
static void interrupt_signal(int signo, siginfo_t *si, void *opaque)
{
	int value;

	if (si->si_code != SI_QUEUE)
		return;

	value = si->si_value.sival_int;

	if (value < INTR_NUM_MIN || value > INTR_NUM_MAX)
		error(1, "wrong interrupt number %d", value);

	debug("receive interrupt %d", value);

	interrupt_state.iflag |= 1 << value;
}

static void cycle_signal(int signo, siginfo_t *si, void *opaque)
{
	int count;

	if (si->si_code != SI_TIMER)
		return;

	debug("receive cycle signal");

	gettimeofday(&tv_begin, NULL);

	count = timer_getoverrun(cycle_timerid);
	if (count > 0) {
		info("cycle timer overruns %d, runtime broke a time condition",
						count);
		hal_runtime_disable();
		return;
	}

	run_cycle_task = 1;
}

static void term_signal(int signo, siginfo_t *si, void *opaque)
{
	switch (signo) {
	case SIGINT:
	case SIGTERM:
		info("exiting by signal %d", signo);
		debug("terminate: %%pc -- 0x%x\n", machine->pc);
		hal_runtime_disable();
		break;
	}
}

/* Add a new task to the interrupt tasks list. */
static void interrupt_add(struct task *itask)
{
	struct task *iter;

	if (inter_list != NULL) {
		for (iter = inter_list; iter->next_task != NULL; iter = iter->next_task)
			;
		iter->next_task = itask;
	} else {
		inter_list = itask;
	}
}

/* Return an interrupt task by its number. */
static struct task *get_inter_by_num(unsigned int intrn)
{
	struct task *iter;

	return_val_if_fail(intrn <= INTR_NUM_MAX, NULL);

	for (iter = inter_list; iter != NULL; iter = iter->next_task) {
		if (iter->intrn == intrn)
			return iter;
	}

	return NULL;
}

/* Helpers to block and unblock a specified signal. */
static void sig_block(int signo)
{
	sigset_t mask;

	sigemptyset(&mask);
	sigaddset(&mask, signo);

	if (sigprocmask(SIG_BLOCK, &mask, NULL) == -1)
		error(1, "sigprocmask failed");
}

static void sig_unblock(int signo)
{
	sigset_t mask;

	sigemptyset(&mask);
	sigaddset(&mask, signo);

	if (sigprocmask(SIG_UNBLOCK,&mask, NULL) == -1)
		error(1, "sigprocmask failed");
}

/* Set a mask for a new task. */
static void change_interrupt_level()
{
	struct task *task;
	uint32_t ilevel;

	task = sched_get_current();
	ilevel = 0;

	if (task == NULL) {
		interrupt_state.ilevel = ~ilevel;
		return;
	}

	switch (task->mode) {
	case TASK_MODE_BACKGROUND:
	case TASK_MODE_CYCLE:
		interrupt_state.ilevel = ~ilevel;
		break;
	case TASK_MODE_INTERRUPT:
		ilevel |= (1 << task->intrn);
		ilevel--;
		interrupt_state.ilevel = ilevel;
		break;
	default:
		error(1, "unknown task type");
	}
}

/* Make a cycle task as a current task. */
static void process_cycle_task()
{
	switch (sched_get_state()) {
	case SCHED_IDLE:
	case SCHED_RUNNING_BG:
		sig_block(SIGCYCLE);
		/* Load cycle task. */
		sched_put_task(cycle_task);
		/* Start images synching. */
		hal_runtime_cycle_begin();
		hal_runtime_cycle_run();
		run_cycle_task = 0;
		break;
	case SCHED_RUNNING_INTR:
	case SCHED_RUNNING_CYCLE:
		error(1, "unexpected scheduler state");
	default:
		break;
	}
}

/* Choose an interrupt task by its numbet and set it as a current task. */
static void process_interrupts(struct vm *machine)
{
	struct task *itask;
	uint32_t iflag;
	unsigned int intrn;

	return_if_fail(machine != NULL);

	sig_block(SIGINTERRUPT);

	iflag = interrupt_state.iflag;
	intrn = 0;
	/* Find the most high prio interrupt. */
	while (!(iflag & 0x01)) {
		iflag >>= 1;
		++intrn;
	}

	debug("got interrupt %d", intrn);
	
	itask = get_inter_by_num(intrn);
	if (itask != NULL)
		sched_put_task(itask);
	else
		warn("no handler for interrupt %d", intrn);

	/* Unset interrupt bit. */
	iflag = 0;
	iflag |= 1 << intrn;
	interrupt_state.iflag &= ~iflag;

	sig_unblock(SIGINTERRUPT);
}

inline static int interrupts_pending()
{
	if (interrupt_state.iflag & interrupt_state.ilevel)
		return 1;

	return 0;
}

inline static int cycle_task_allowed()
{
	uint32_t ilevel;

	if (!run_cycle_task)
		return 0;

	ilevel = ~0;
	/* With ilevel we make sure that we don't stop interrupt
	 * task execution. ilevel is ~0 when we are not in  the
	 * interrupt handling. */
	if (interrupt_state.ilevel & ilevel)
		return 1;

	debug("not a cycle time");

	return 0;
}

/* Helpers for managing a cycle timer. */
static void cycle_timer_arm()
{
	struct itimerspec its;
	int res;

	debug("arm cycle timer sec %lu nsec %lu", cycle_timeout_sec, cycle_timeout_nsec);
	
	its.it_value.tv_sec = cycle_timeout_sec;
	its.it_value.tv_nsec = cycle_timeout_nsec;
	its.it_interval.tv_sec = its.it_value.tv_sec;
	its.it_interval.tv_nsec = its.it_value.tv_nsec;

	res = timer_settime(cycle_timerid, CLOCK_REALTIME, &its, NULL);
	if (res == -1)
		error(1, "timer_settime failed: %s", strerror(errno));
}

static void cycle_timer_disarm()
{
	struct itimerspec its;
	int res;

	debug("disarm cycle timer");

	/* Disarm timer conditions. */
	its.it_value.tv_sec = 0;
	its.it_value.tv_nsec = 0;
	its.it_interval.tv_sec = 0;
	its.it_interval.tv_nsec = 0;

	res = timer_settime(cycle_timerid, CLOCK_REALTIME, &its, NULL);
	if (res == -1)
		error(1, "timer_settime failed: %s", strerror(errno));
}

static void cycle_timer_create()
{
	struct sigevent sev;
	int res;

	memset(&sev, 0, sizeof(struct sigevent));
	sev.sigev_notify = SIGEV_SIGNAL;
	sev.sigev_signo = SIGCYCLE;

	res = timer_create(CLOCK_REALTIME, &sev, &cycle_timerid);
	if (res == -1)
		error(1, "timer_create failed: %s", strerror(errno));

	cycle_timer_arm();
}

static void cycle_timer_delete()
{
	int res;

	cycle_timer_disarm();

	res = timer_delete(cycle_timerid);
	if (res == -1)
		error(1, "timer_delete failed");
}

static void realtime_enter(int priority)
{
	struct sched_param parameters;
	int res;

	parameters.sched_priority = priority;
	res = sched_setscheduler(0, SCHED_RR, &parameters);
	if (res) {
		info("Failed to enter RT SCHED_RR scheduler: %s", strerror(errno));
                return;
        }

	debug("RT scheduler priority %d", priority);
}

static void compute_cycle_stats()
{
	struct timeval tv_res;
	unsigned long cycle_usec;

	gettimeofday(&tv_end, NULL);
	timersub(&tv_end, &tv_begin, &tv_res);
	
	cycle_usec = tv_res.tv_sec * 1000000 + tv_res.tv_usec;

	if (vmachined_stats.cycle_usec_min) {
		if (cycle_usec < vmachined_stats.cycle_usec_min)
			vmachined_stats.cycle_usec_min = cycle_usec;
	} else {
		vmachined_stats.cycle_usec_min = cycle_usec;
	}

	if (vmachined_stats.cycle_usec_max) {
		if (cycle_usec > vmachined_stats.cycle_usec_max)
			vmachined_stats.cycle_usec_max = cycle_usec;
	} else {
		vmachined_stats.cycle_usec_max = cycle_usec;
	}

	vmachined_stats.cycle_usec_last = cycle_usec;
	vmachined_stats.cycle_run++;
}

/* If nothing to do for our machine the process sleeps. */
static void sleep_until_interrupted()
{
	int res;

	debug("I gonna to bed");
	
	res = poll(NULL, 0, -1);
	if (res == -1 && errno != EINTR)
		error(1, "poll failed");
}

/* Handlers for the hal-runtime core. */
static void connect_reply_cb(unsigned int inputs_sz, unsigned int outputs_sz)
{
	debug("connect reply: inputs size %d, outputs size %d allocate images",
				inputs_sz, outputs_sz);

	if (inputs_sz != 0)  {
		inputs_image = xmalloc0(inputs_sz);
		inputs_image_sz = inputs_sz;
	}

	if (outputs_sz != 0) {
		outputs_image = xmalloc0(outputs_sz);
		outputs_image_sz = outputs_sz;
	}

	vm_io_area_set(machine, inputs_image, inputs_sz,
		       outputs_image, outputs_image_sz, byte_order);
}

static void cycle_begin_reply_cb()
{
	debug("cycle begin reply: we have valid images");
}

static void cycle_run_reply_cb()
{
	debug("cycle run reply: now we start machine");
}

static void cycle_end_reply_cb()
{
	debug("cycle end reply: images were sent");
}

static void disconnect_reply_cb()
{
	debug("disconnect reply: bye bye baby");
}

static const char *write_inputs_request_cb(const char inputs[HAL_IMAGE_MAX+1],
					   unsigned int off)
{
	unsigned char *p, n;
	unsigned int len, i;

	len = strlen(inputs);
	assert(len % 2 == 0);

	len >>= 1;

	debug("write inputs request: inputs %s offset %d", inputs, off);

	if (len + off > inputs_image_sz)
		return "Inputs image len out of the range";

	/* Put data to image. Ascii to hex transformation. */
	p = inputs_image;
	p += off;
	len <<= 1;

	for (i = 0; i < len; i+= 2) {
		n = ascii_to_hex(inputs[i+1]);
		n += ascii_to_hex(inputs[i]) << 4;
		*p++ = n;
	}

	return HAL_OK;
}

static const char *write_outputs_request_cb(const char outputs[HAL_IMAGE_MAX+1],
					    unsigned int off)
{
	unsigned char *p, n;
	unsigned int len, i;

	len = strlen(outputs);
	assert(len % 2 == 0);

	len >>= 1;

	debug("write outputs request: outputs %s offset %d", outputs, off);

	if (len + off > outputs_image_sz)
		return "Outputs image len out of the range";

	/* Put data to image. Ascii to hex transformation. */
	p = outputs_image;
	p += off;
	len <<= 1;

	for (i = 0; i < len; i+= 2) {
		n = ascii_to_hex(outputs[i+1]);
		n += ascii_to_hex(outputs[i]) << 4;
		*p++ = n;
	}

	return HAL_OK;
}

static const char *read_inputs_request_cb(char inputs[HAL_IMAGE_MAX+1],
					  unsigned int len, unsigned int off)
{
	unsigned char *p;
	unsigned int i;

	debug("read inputs request: lenght %d offset %d", len, off);

	if (len + off > inputs_image_sz)
		return "Inputs image size out of the range";

	p = inputs_image;
	p += off;

	for (i = 0; i < len; i++) {
		inputs[2*i] = hex_to_ascii((p[i] >> 4) & 0x0f);
		inputs[2*i+1] = hex_to_ascii(p[i] & 0x0f);
	}

	inputs[2*len] = 0;

	return HAL_OK;
}

static const char *read_outputs_request_cb(char outputs[HAL_IMAGE_MAX+1],
					   unsigned int len, unsigned int off)
{
	unsigned char *p;
	unsigned int i;

	debug("read outputs request: lenght %d offset %d", len, off);

	if (len + off > outputs_image_sz)
		return "Outputs image size out of the range";

	p = outputs_image;
	p += off;

	for (i = 0; i < len; i++) {
		outputs[2*i] = hex_to_ascii((p[i] >> 4) & 0x0f);
		outputs[2*i+1] = hex_to_ascii(p[i] & 0x0f);
	}

	outputs[2*len] = 0;

	return HAL_OK;
}

static void print_usage()
{
	printf("Usage: %s [-ih] [-c config]\n"
	" -i		stay interactive\n"
	" -h		print this help\n"
	" -c config	path to a config file\n", __progname);
}

int main(int argc, char **argv)
{
	struct vm_options opts;
	struct hal_runtime_callbacks callbacks;
	char filepath[PATH_MAX];
	char configfile[PATH_MAX];
	char *path_to_lib, *ptr;
	struct task_conf *iter_conf;
	struct task *task, *iter_task;
	unsigned int i;
	int res, sfd, cflag;
       	int priority, opt, no_detach;
	ret_t ret;
	
	/* Now it is always true. We don't have a log subsystem. */
	no_detach = 1;
	cflag = 0;

	while ((opt = getopt(argc, argv, "ihc:")) > 0) {
		switch (opt) {
		case 'i':
			no_detach = 1;
			break;
		case 'h':
			print_usage();
			exit(0);
		case 'c':
			xstrncpy(filepath, optarg, sizeof(filepath));
			ptr = canonize_path(filepath);
			xstrncpy(configfile, ptr, sizeof(configfile));
			xfree(ptr);
			cflag = 1;
			break;
		default:
			print_usage();
			exit(-1);
		}
	}

	argc -= optind;
	argv += optind;

	if (!cflag)
		error(1, "config file option not found");

	res = read_config(configfile, &opts);
	if (res != 0)
		error(1, "can't parse a machine configuration file");

	debug("working directory: %s", opts.work_dir);
	debug("library call directory: %s", opts.libcall_dir);
	debug("project directory: %s", opts.exec_dir);

	machine = vm_create();

	asprintf(&path_to_lib, "%s/%s", opts.work_dir, opts.libcall_dir);
	ret = vm_load_lib(machine, path_to_lib);
	if (ret != ret_ok)
		error(1, "can't load library");

	free(path_to_lib);

	/* Timeout value in miliseconds converts into seconds and
	 * nanoseconds. */
	if (opts.cycle_timeout > 0) {
		cycle_timeout_sec = opts.cycle_timeout / 1000;
		cycle_timeout_nsec = opts.cycle_timeout * 1000000 % 1000000000;
	} else {
		cycle_timeout_sec = CYCLE_TIMEOUT_SEC_DEFAULT;
		cycle_timeout_nsec = 0;
	}

	vmachined_stats.cycle_usec_user = cycle_timeout_sec * 1000000 + cycle_timeout_nsec / 1000;

	priority = VMACHINED_RT_PRIORITY;

	if (opts.priority > 0) {
		if (opts.priority < sched_get_priority_min(SCHED_RR) || 
		    opts.priority > sched_get_priority_max(SCHED_RR)) 
			error(1, "Priority range is %d..%d",
				sched_get_priority_min(SCHED_RR),
				sched_get_priority_max(SCHED_RR));
		else
			priority = opts.priority;
	}

	if (opts.byte_order > 0) 
		byte_order = opts.byte_order;
	else
		byte_order = BO_NATIVE;

	if (opts.tasks == NULL)
		error(1, "no tasks");

	for (iter_conf = opts.tasks; iter_conf != NULL; iter_conf = iter_conf->next) {
		char *file_path;

		debug("");
		debug("task file name %s", iter_conf->file_name);
		debug("task entry point %s", iter_conf->task_name);
		if (iter_conf->mode == TASK_MODE_INTERRUPT)
			debug("interrupt number %d", iter_conf->intrn);
		debug("");

		asprintf(&file_path, "%s/%s/%s", opts.work_dir,
						opts.exec_dir, iter_conf->file_name);
		ret = vm_load_file(machine, file_path, iter_conf->task_name, 1);
		if (ret != ret_ok)
			error(ret, "can't load file %s", file_path);
		free(file_path);

		ret = vm_prepare(machine);
		if (ret != ret_ok)
			error(ret, "machine is not prepared");

		task = task_new(iter_conf->mode, iter_conf->task_name);
		task_get_ctx(machine, task);
		debug("Task %s: pc -- 0x%x", task->name, task->pc);

		switch (iter_conf->mode) {
		case TASK_MODE_INTERRUPT:
			debug("inter: %s", task->name);
			task->intrn = iter_conf->intrn;
			interrupt_add(task);
			break;
		case TASK_MODE_CYCLE:
			debug("cycle: %s", task->name);
			cycle_task = task;
			break;
		case TASK_MODE_BACKGROUND:
			debug("bg: %s", task->name);
			bg_task = task;
			break;
		default:
			error(1, "unknown task mode");
		}
	}

	sigevent_ignore(SIGPIPE);
	sigevent_add(SIGINT, term_signal, NULL, 1);
	sigevent_add(SIGTERM, term_signal, NULL, 1);
	sigevent_add(SIGINTERRUPT, interrupt_signal, NULL, 1);
	sigevent_add(SIGCYCLE, cycle_signal, NULL, 1);
	sigevent_init();

	timeout_init();
	io = io_new();

	if (strlen(opts.socket) != 0) {
		sfd = sock_unix_connect(opts.socket);
		if (sfd == -1)
			error(1, "can't connect to %s", opts.socket);
	} else if (strlen(opts.inet_addr) != 0) {
		char *p;
		p = strchr(opts.inet_addr, ':');
		if (p == NULL)
			error(1, "can't find a host/port separation character");

		*p++ = 0;
		sfd = sock_inet_connect_wait("tcp", "ip", opts.inet_addr, p, 0);
		if (sfd == -1)
			error(1, "can't connect to %s:%s", opts.inet_addr, p);
	}
	
	hal_runtime_init(io, sfd, sfd, 0);

	memset(&callbacks, 0, sizeof(callbacks));

	callbacks.connect_callback = connect_reply_cb;
	callbacks.disconnect_callback = disconnect_reply_cb;
	callbacks.cycle_begin_callback = cycle_begin_reply_cb;
	callbacks.cycle_run_callback = cycle_run_reply_cb;
	callbacks.cycle_end_callback = cycle_end_reply_cb;
	callbacks.disconnect_callback = disconnect_reply_cb;
	callbacks.write_inputs_callback = write_inputs_request_cb;
	callbacks.write_outputs_callback = write_outputs_request_cb;
	callbacks.read_inputs_callback = read_inputs_request_cb;
	callbacks.read_outputs_callback = read_outputs_request_cb;

	hal_runtime_register_callbacks(&callbacks);
	hal_runtime_connect();

	realtime_enter(priority);

	if (bg_task != NULL)
		sched_put_task(bg_task);

	cycle_timer_create();

	if (!no_detach)
		daemonize();

	while (hal_runtime_state() != HAL_RT_STATE_DISABLED &&
	       hal_runtime_state() != HAL_RT_STATE_ERROR) {
		/* Choose a new task and load machine with this task. */
		sched_tasks(machine);
		/* Change ilevel according to a new task. */
		change_interrupt_level();

		if (sched_get_state() == SCHED_IDLE)
			sleep_until_interrupted();
		else
			debug("task %s start with pc 0x%x",
				sched_get_current()->name, machine->pc);

		/* Execute task till not interrupts, cycle timer or the end of
		 * the current task. */
		while (!interrupts_pending() && !cycle_task_allowed() &&
				sched_get_state() != SCHED_IDLE) {
			vm_exec_instr(machine);
			if (machine->state == VM_STOPPED) {
				debug("task %s is over", sched_get_current()->name);

				switch (sched_get_state()) {
				case SCHED_RUNNING_BG:
					/* Background task is always loaded.
					 * Don't waste time for scheduler.  */
					task_switch_ctx(machine, bg_task, bg_task);
					break;
				case SCHED_RUNNING_CYCLE:
					/* Come to idle state. */
					sched_current_del(machine);
					/* Sync images with hal manager. */
					hal_runtime_cycle_end();
					/* Do cycle stats. */
					compute_cycle_stats();
					/* Unblock cycle signal. */
					sig_unblock(SIGCYCLE);
					break;
				case SCHED_RUNNING_INTR:
					/* Come to idle state. */
					sched_current_del(machine);
					break;
				default:
					error(1, "unexpected scheduler state");
				}
			}
		}
		
		if (run_cycle_task && !interrupts_pending()) {
			process_cycle_task();
			continue;
		}

		if (interrupts_pending())
			process_interrupts(machine);
	}

	cycle_timer_delete();

	/* In case of the error state we can reinit a connection to hal-mgr
	 * and go back to cycle. Think about that we can jump back from
	 * here.
	 */

	hal_runtime_deinit();
	io_free(&io);
	timeout_deinit();
	sigevent_deinit();

	iter_task = inter_list;

	while (iter_task != NULL) {
		struct task *tmp;

		task_set_ctx(machine, iter_task);
		ret = vm_unload_file(machine);
		if (ret != ret_ok)
			error(1, "can't unload file");

		tmp = iter_task;
		iter_task = iter_task->next_task;
		xfree(tmp);
	}

	if (bg_task != NULL) {
		task_set_ctx(machine, bg_task);
		ret = vm_unload_file(machine);
		if (ret != ret_ok)
			error(1, "can't unload file");
		xfree(bg_task);
	}

	if (cycle_task != NULL) {
		task_set_ctx(machine, cycle_task);
		ret = vm_unload_file(machine);
		if (ret != ret_ok)
			error(1, "can't unload file");
		xfree(cycle_task);
	}

	vm_io_area_unset(machine);

	ret = vm_unload_lib(machine);
	if (ret != ret_ok)
		error(1, "can't unload library");
		
	ret = vm_destroy(&machine);
	if (ret != ret_ok)
		error(1 ,"can't destroy machine");
	
	iter_conf = opts.tasks;

	while (iter_conf != NULL) {
		struct task_conf *tmp;
		tmp = iter_conf;
		iter_conf = iter_conf->next;
		free(tmp);
	}

	info("");
	info("Perfomance counters:");
	info("");
	info("cycle run %lu times", vmachined_stats.cycle_run);
	info("cycle usec last %lu", vmachined_stats.cycle_usec_last);
	info("cycle usec min %lu", vmachined_stats.cycle_usec_min);
	info("cycle usec max %lu", vmachined_stats.cycle_usec_max);
	info("cycle usec user %lu", vmachined_stats.cycle_usec_user);

	info("");
	info("inputs image content:");
	for (i = 0; i < inputs_image_sz; i++)
		info("%02x", inputs_image[i]);

	info("");
	info("outputs image content:");
	for (i = 0; i < outputs_image_sz; i++)
		info("%02x", outputs_image[i]);

	if (inputs_image != NULL)
		xfree(inputs_image);

	if (outputs_image != NULL)
		xfree(outputs_image);

	info("");
	info("The end sweety");
	info("");

	return 0;
}

